library nueli.ui.screen;

import 'dart:async';

import 'package:flutter/cupertino.dart';
import 'package:flutter/material.dart';
import 'package:flutter/services.dart';
import 'package:nueli_core/nueli_core.dart';

import 'ui.dart';

/// Screen
abstract class Screen extends UI {
  final GlobalKey<ScaffoldState> driver = GlobalKey<ScaffoldState>();
  late final Symbol id;
  late final ScreenRoute route;

  Screen({Key? key, Symbol? id}) : super(key: key) {
    this.id = id ?? ID.newInScope;
    route = ScreenRoute(this);
  }

  Decor get displayBarrier {
    return $decor
      ..onTap(() => close)
      ..bucketFill(palette.shade);
  }

  Decor get displayBackground => $decor;

  @mustCallSuper
  @protected
  Widget get displayWrapper {
    return Scaffold(
      key: driver,
      backgroundColor: palette.none,
      drawer: drawer,
      endDrawer: endDrawer,
      body: Stack(children: [
        $layout..children([header, body]),
        ...floatingLayers,
        floatingHeader,
      ]),
    );
  }

  Widget get header => empty;

  Widget get body;

  Widget? get drawer => null;

  Widget? get endDrawer => null;

  Widget get floatingHeader => empty;

  List<Widget> get floatingLayers => [];

  Tranx get tranx => $tranx.translate(const Offset(0, 1), Offset.zero).fade(0, 1);

  Tranx get secondaryTranx => $tranx;

  @mustCallSuper
  void close() => Screens.of(context).goBack();

  @mustCallSuper
  void openDrawer() => driver.currentState!.openDrawer();

  @mustCallSuper
  void openEndDrawer() => driver.currentState!.openDrawer();

  @override
  Widget call() {
    return Material(
      color: palette.none,
      child: Stack(children: [
        $tranx.animation(route.animation).decor($decor.sizeToInfinity.onTap(close), displayBarrier),
        displayBackground.sizeToInfinity,
        tranx.animation(route.animation).child(
              secondaryTranx.animation(route.secondaryAnimation).child(displayWrapper),
            )
      ]),
    );
  }

  @override
  ScreenState<Screen> createState() => ScreenState();
}

class ScreenState<T extends Screen> extends UIState<T> {
  Future<bool> Function()? backPressCallback;

  ScreenState([this.backPressCallback]);

  List<Logic> get logicArray => [];

  @mustCallSuper
  Future<dynamic> initAsync() async {
    await LogicTree.compile(logicArray);
  }

  @override
  Widget build(BuildContext context) {
    return WillPopScope(
      onWillPop: backPressCallback,
      child: super.build(context),
    );
  }
}

/// Screen Route
class ScreenRoute extends ModalRoute<ScreenRoute> {
  final Screen screen;

  ScreenRoute(this.screen);

  @override
  RouteSettings get settings => RouteSettings(name: screen.id.name);

  @override
  @mustCallSuper
  Duration get transitionDuration => screen.tranx().duration;

  @override
  @mustCallSuper
  Duration get reverseTransitionDuration {
    return screen.secondaryTranx().reverseDuration;
  }

  @override
  @mustCallSuper
  bool get maintainState => true;

  @override
  @mustCallSuper
  bool get opaque => true;

  @override
  @mustCallSuper
  String get barrierLabel => '';

  @override
  @mustCallSuper
  Color? get barrierColor => palette.none;

  @override
  @mustCallSuper
  bool get barrierDismissible => false;

  @override
  @mustCallSuper
  Widget buildPage(
    BuildContext context,
    Animation<double> animation,
    Animation<double> secondaryAnimation,
  ) {
    return screen;
  }

  @override
  @mustCallSuper
  Widget buildTransitions(
    BuildContext context,
    Animation<double> animation,
    Animation<double> secondaryAnimation,
    Widget child,
  ) {
    super.buildTransitions(context, animation, secondaryAnimation, child);
    return screen;
  }
}

/// ScreenApp
class ScreenApp extends StatelessWidget {
  final Symbol id;
  final Screen entry;
  final Future<bool> Function()? exitCallBack;
  final bool showDebugSemantics;
  final bool showDebugBanner;
  final Function()? init;

  const ScreenApp({
    Key? key,
    this.id = #general,
    required this.entry,
    this.exitCallBack,
    this.showDebugBanner = false,
    this.showDebugSemantics = false,
    this.init,
  }) : super(key: key);

  @override
  Widget build(BuildContext context) {
    (init ?? () {})();
    return MaterialApp(
      debugShowCheckedModeBanner: showDebugBanner,
      showSemanticsDebugger: showDebugSemantics,
      home: entry,
      color: palette.background,
      theme: theme,
    );
  }

  static ThemeData get theme => ThemeData(
        primaryColor: palette.primary,
        primaryColorDark: palette.primaryDark,
        backgroundColor: palette.background,
        scaffoldBackgroundColor: palette.background,
        colorScheme: ColorScheme.fromSwatch(
          primarySwatch: palette.material as MaterialColor,
          primaryColorDark: palette.primaryDark,
          accentColor: palette.accent,
          backgroundColor: palette.background,
          cardColor: palette.background,
          errorColor: palette.error,
        ).copyWith(
          primary: palette.primary,
          secondary: palette.accent,
          background: palette.background,
          error: palette.error,
          onPrimary: palette.onPrimary,
          onSecondary: palette.onAccent,
          onBackground: palette.onBackground,
          onError: palette.background,
        ),
      );
}

/// Screens
class Screens {
  final BuildContext context;

  Screens.of(this.context);

  open(ScreenRoute screen) {
    return Navigator.push(context, screen);
  }

  openReplacement(ScreenRoute screen) {
    return Navigator.pushReplacement(context, screen);
  }

  openRollback(ScreenRoute screen) {
    return Navigator.pushAndRemoveUntil(context, screen, (r) => false);
  }

  bool get canGoBack => Navigator.of(context).canPop();

  void goBack() {
    if (canGoBack) return Navigator.of(context).pop();
  }

  goBackTo(ScreenRoute screen) {
    Navigator.of(context).popUntil((route) => route == screen);
  }

  void goHome() {
    Navigator.of(context).popUntil((route) => route.isFirst);
  }

  void exit() {
    SystemNavigator.pop(animated: true);
  }
}

/// Logic
abstract class Logic {
  Symbol get channel;
  Algorithm get initializer;
  Algorithm get reviver => algog;
}

/// Logic Tree
class LogicTree {
  static final Map<Symbol, dynamic> _channelBus = {};

  static Future<void> compile(List<Logic> logicArray) async {
    WidgetsFlutterBinding.ensureInitialized();
    for (Logic logic in logicArray) {
      final hasInit = !_channelBus.containsKey(logic.channel);
      _channelBus.putIfAbsent(logic.channel, () => 'init');
      hasInit ? await logic.initializer() : await logic.reviver();
    }
  }
}
